module Ex03 where

import Data.List
import Test.QuickCheck

-- Consider the following six properties of the nub function:
--
-- 1. nub (nub xs) == nub xs
-- 2. nub (x : nub xs) == nub (x : xs)
-- 3. nub (xs ++ nub ys) == nub (xs ++ ys)
-- 4. nub (x : x : xs) == nub (x : xs)
-- 5. nub (xs ++ [x] ++ xs) = nub (xs ++ [x])
-- 6. nub [x] == [x]

e1 :: ([Int] -> [Int]) -> [Int] -> Bool
e1 f xs = f (f xs) == f xs
-- Idempotence law, f^2 = f
-- "sort", id, constants
-- reverse, remove every second element, flipping the first two elements etc.

e2 :: ([Int] -> [Int]) -> Int -> [Int] -> Bool
e2 f x xs = f (x : f xs) == f (x : xs)
-- Idempotence type law
-- id, constants, filter (with some condition)
-- reverse, remove every second element, flipping the first two elements etc.

e3 :: ([Int] -> [Int]) -> [Int] -> [Int] -> Bool
e3 f xs ys = f (xs ++ f ys) == f (xs ++ ys)
-- Idempotence law
-- id, constants, filter ... 

e4 :: ([Int] -> [Int]) -> Int -> [Int] -> Bool
e4 f x xs = f (x : x : xs) == f (x : xs)
--duplicate rule
-- satisfied by: nub, constant, take 1, take 1 . reverse
-- Broken by: id, reverse, duplicate first element etc.

-- [] ++ [1] ++ [2, 3]
-- [1, 2, 3, 5, 3, 2, 1]
e5 :: ([Int] -> [Int]) -> Int -> [Int] -> Bool
e5 f x xs = f (xs ++ [x] ++ xs) == f (xs ++ [x])
-- duplicate rule
-- Satisfied and broken by many above
-- Reordering ought to break

e6 :: ([Int] -> [Int]) -> Int -> Bool
e6 f x = f [x] == [x]
-- id on singletons
-- rules out constants


dnub1236 :: (Eq a) => [a] -> [a]
dnub1236 xs = xs

-- id (x : x : xs) == id (x : xs)
dnub1236_c4 :: (Int, [Int]) -- (x,xs) failing eqn 4
dnub1236_c4 = (5, [6, 7])

-- f (xs ++ [x] ++ xs) == f (xs ++ [x])
dnub1236_c5 :: (Int, [Int]) -- (x,xs) failing eqn 5
dnub1236_c5 = (0, [0])

-- Task 2. Implement a dodgy nub function that satisfies
-- equations 1, 2, 3, 4, 5, but not 6.
-- Give counterexamples.

dnub12345 :: (Eq a) => [a] -> [a]
dnub12345 xs = []

-- f [x] == [x]
dnub12345_c6 :: Int
dnub12345_c6 = 5

-- Task 3. Implement a dodgy nub function that satisfies
-- equations 4, 5, 6, but not 1,2,3.
-- Give counterexamples.

-- This works with essentially any non-idempotent function
-- For example, you could replace reverse with a function
-- that just swaps the first two elements of a list (if there
-- exist such elements to swap) and is id on singletons.

dnub456 :: (Eq a) => [a] -> [a]
dnub456 xs = reverse $ nub xs

-- f (f xs) == f xs, [1, 2] => [2, 1]
dnub456_c1 :: [Int]
dnub456_c1 = [1, 2]

-- f (x : f xs) == f (x : xs), f 1 : (f [1, 2]) == f [1, 2, 1] == [1, 2] /= [1, 1, 2] => [2, 1]
dnub456_c2 :: (Int,[Int])
dnub456_c2 = (1, [1, 2])

-- f (xs ++ f ys) == f (xs ++ ys)
-- f (3, 4, 6, 5) = [5, 6, 4, 3]
-- f (3, 4, 5, 6) = [6, 5, 4, 3]
dnub456_c3 :: ([Int],[Int]) -- (xs,ys)
dnub456_c3 = ([3, 4], [5, 6])

-- Task 4. Implement a dodgy nub function that satisfies
-- all the equations 1,2,3,4,5,6. (3 marks)

dnub123456 :: (Eq a) => [a] -> [a]
dnub123456 xs = take 1 xs

-- Another valid implementation for dnub123456,
-- which uses a similar idea to what we did for
-- dnub456
dnub123456' xs = let ys = nub xs
               in if ys /= reverse ys then concatMap (\x -> [x, x]) ys else ys


-- Show that dnub123456 is dodgy, i.e. give an input xs for which
-- dnub123456 xs /= nub xs.
-- This works for both dnub123456 and dnub123456'
dnub123456_c :: [Int]
dnub123456_c = [0, 1]


