module SearchTree (BST, isBST, treeInsert, emptyBST, traverseTree) where
-- Define a module SearchTree which exports the below functions 
import Test.QuickCheck
import Data.List
-- Define a data type BST representing a binary search tree with Integer nodes
data BST = Leaf | Branch Integer BST BST deriving (Show, Ord, Eq)

-- Define a function isBST which checks whether a BST is a lawful BST
isBST :: BST -> Bool
isBST Leaf = True
isBST (Branch v l r) = allTree (< v) l && allTree (>= v) r && isBST l && isBST r where
    allTree :: (Integer -> Bool) -> BST -> Bool
    allTree f (Branch u l r) = f u && allTree f l && allTree f r
    allTree f Leaf = True
     
-- Implement the insert function, which takes an integer to a BST, preserving BST property.
treeInsert :: Integer -> BST -> BST
treeInsert i Leaf = Branch i Leaf Leaf
treeInsert i (Branch v l r) | i < v     = Branch v (treeInsert i l) r
                            | otherwise = Branch v l (treeInsert i r)

emptyBST :: BST
emptyBST = Leaf

traverseTree :: BST -> [Integer]
traverseTree Leaf = []
traverseTree (Branch v l r) = traverseTree l ++ [v] ++ traverseTree r


-- This was written outside the practical, so don't worry if it's not understandable
-- It's just so we have a way of generating arbitrary trees
-- It will be clear soon what this is saying, when we cover monads and do notation
searchTrees :: Gen BST
searchTrees = sized searchTrees'
  where 
   searchTrees' 0 = return Leaf
   searchTrees' n = do 
      v <- (arbitrary :: Gen Integer)
      fmap (treeInsert v) (searchTrees' $ n - 1)

instance Arbitrary BST where
    arbitrary = searchTrees